from math import *
from euclid import *
from omega import *
from cyclops import *
import time
import utm

scene = getSceneManager()
cam = getDefaultCamera()
mainMenu = MenuManager.createAndInitialize()

mainLight = Light.create()
mainLight.setLightType(LightType.Directional)
mainLight.setLightDirection(Vector3(-1.0, -1.0, -1.0))
mainLight.setColor(Color(0.5, 0.5, 0.5, 1.0))
mainLight.setAmbient(Color(0.25, 0.25, 0.25, 1.0))
mainLight.setEnabled(True)

# Referenced by crime type menu
crimeTypeSelection = "All"

# Referenced by the time index menus
crimeStartYear = 2001
crimeStartMonth = 1
crimeEndYear = 2013
crimeEndMonth = 8

# --------------------------------------------
# Load in the OSG, Google, and Yahoo "models"
# --------------------------------------------
osgModel = ModelInfo()
osgModel.name = "osgCity"
osgModel.path = "chicago_flat.earth"
scene.loadModel(osgModel)
osgCity = StaticObject.create("osgCity")
osgCity.getMaterial().setLit(False)

googleModel = ModelInfo()
googleModel.name = "googleCity"
googleModel.path = "chicago_flat.map.earth"
scene.loadModel(googleModel)
googleCity = StaticObject.create("googleCity")
googleCity.getMaterial().setLit(False)

yahooModel = ModelInfo()
yahooModel.name = "yahooCity"
yahooModel.path = "chicago_yahoo.earth"
scene.loadModel(yahooModel)
yahooCity = StaticObject.create("yahooCity")
yahooCity.getMaterial().setLit(False)

# ------------------------
# General initializations
# ------------------------
def seeCity():
	camOffset = osgCity.getBoundRadius()
	cam.setPosition(osgCity.getBoundCenter()+Vector3(0.0, -camOffset, camOffset))
	cam.setPitchYawRoll(Vector3(math.pi*0.25, 0.0, 0.0))
seeCity()

cam.getController().setSpeed(10000)
setNearFarZ(1.0, 2.0*osgCity.getBoundRadius())
queueCommand(':autonearfar on') # Alessandro's suggestion for the triangles

# ------------------------------------------------------------------------------------------------
# Read in the community area data (areas and crimes) into allCommunities[], an array of Community
# ------------------------------------------------------------------------------------------------
class Crime:
	def __init__(self, inDate, inKind):
		self.year = int(inDate[:4])
		self.month = int(inDate[4:6])
		self.kind = inKind

class Community:
	num = 0
	def __init__(self, inName, inPolyCenter, inBoundary):
		self.name = inName
		self.polyCenter = inPolyCenter
		self.boundary = inBoundary
		self.crimes = []
		self.surfaceTop = None
		self.surfaceSide = None
		Community.num += 1

allCommunities = []
f = open("communityAreas.csv")
for line in f:
	entries = line.split(',')
	entries[-1] = entries[-1].rstrip()
	name = entries.pop(0)
	
	start = utm.from_latlon(float(entries[1]), float(entries[0])) # Yes, the the CSV has them backwards
	boundaryVertices = [Vector2(start[0], start[1])]
	boundaryCenter = Vector2(start[0], start[1])
	
	eLen = len(entries)
	for i in range(2, eLen, 2):
		end = utm.from_latlon(float(entries[i+1]), float(entries[i])) # Yes, the the CSV has them backwards
		
		# Limit how many vertices are used
		if math.hypot(start[0]-end[0], start[1]-end[1]) < 30.0: continue
		
		# The center of the boundary will be needed later for rendering accuracy
		boundaryVertices.append(Vector2(start[0], start[1]))
		boundaryCenter.x += start[0]
		boundaryCenter.y += start[1]
		start = end
	
	# Average the boundary vertices
	boundaryCenter /= len(boundaryVertices)
	
	allCommunities.append(Community(name, boundaryCenter, boundaryVertices))
f.close()

ticks = time.time()
f = open("reducedComCrimes.csv")
for i, line in enumerate(f):
	entries = line.split(',')
	entries[-1] = entries[-1].rstrip()
	
	# Associate a crime with a community ...this was pre-calculated with ~400,000,000 point in polygon tests
	#if (int(entries[1][:4]) >= 2011):
	allCommunities[int(entries[0])].crimes.append(Crime(entries[1], entries[2]))
	
	if i != 0 and i%100000 == 0: print str(i)+": "+entries[0]+", "+str(time.time()-ticks)
f.close()
print "Time to read all community crimes = "+str(time.time()-ticks)

# --------------------------------------------------------------------------
# Read in both CTA data (lines and stations) into allCTA[], an array of CTA
# --------------------------------------------------------------------------
class Station:
	def __init__(self, inName, inAddress, inLoc):
		self.name = inName
		self.address = inAddress
		self.loc = inLoc

class CTA:
	num = 0
	def __init__(self, inName, inColor):
		self.name = inName
		self.color = inColor
		self.lineSegments = []
		self.stations = []
		CTA.num += 1

dictCTA = {"Blue":0, "Brown":1, "Green":2, "Orange":3, "Pink":4, "Purple":5, "Red":6, "Yellow":7}
allCTA = [
	CTA("Blue", "colored -d #0000ff -e #000080"),
	CTA("Brown", "colored -d #804000 -e #402000"),
	CTA("Green", "colored -d #00ff00 -e #008000"),
	CTA("Orange", "colored -d #ff4000 -e #802000"),
	CTA("Pink", "colored -d #ff00c0 -e #800060"),
	CTA("Purple", "colored -d #ff00ff -e #800080"),
	CTA("Red", "colored -d #ff0000 -e #800000"),
	CTA("Yellow", "colored -d #ffff00 -e #808000")
]

f = open("ctaLines.csv")
for line in f:
	entries = line.split(',')
	entries[-1] = entries[-1].rstrip()
	colors = entries.pop(0).split(';')
	
	vertices = []
	eLen = len(entries)
	for i in range(0, eLen, 2):
		utmConvert = utm.from_latlon(float(entries[i+1]), float(entries[i])) # Yes, the the CSV has them backwards
		vertices.append(Vector2(utmConvert[0], utmConvert[1]))
	
	indices = []
	for c in colors: indices.append(dictCTA[c])
	for i in indices: allCTA[i].lineSegments.append(vertices)
f.close()

f = open("ctaStations.csv")
for line in f:
	entries = line.split(',')
	entries[-1] = entries[-1].rstrip()
	colors = entries.pop(0).split(';')
	utmConvert = utm.from_latlon(float(entries[3]), float(entries[2])) # Yes, the the CSV has them backwards
	
	indices = []
	for c in colors: indices.append(dictCTA[c])
	for i in indices: allCTA[i].stations.append(Station(entries[0], entries[1], Vector2(utmConvert[0], utmConvert[1])))
f.close()

# -----------------------------------------------------
# Create geometry for each element of allCommunities[]
# -----------------------------------------------------
class Triangle:
	def __init__(self, inA, inB, inC):
		self.a = inA
		self.b = inB
		self.c = inC

twoPI = math.pi*2.0 # This gets used many times in the function below

def triangularize(inBoundary):
	copyBoundary = inBoundary[:] # Don't modify inBoundary, Python is stupid
	outSurface = []
	
	index = 0;
	while len(copyBoundary) > 2:
		cbLen = len(copyBoundary)
		pIndex = (index+cbLen-1)%cbLen;
		nIndex = (index+1)%cbLen;
		
		bp = copyBoundary[pIndex]
		bi = copyBoundary[index]
		bn = copyBoundary[nIndex]
		
		# This assumes the polygon is in clockwise order
		theta = math.atan2(bi.y-bn.y, bi.x-bn.x)-math.atan2(bi.y-bp.y, bi.x-bp.x);
		if theta < 0.0: theta += twoPI;
		
		# If bp, bi, and bn describe an "ear" of the polygon
		if theta < math.pi:
			inside = False;
			
			# Make sure other vertices are not inside the "ear"
			for i in range(cbLen):
				if i == pIndex or i == index or i == nIndex: continue;
				
				# Black magic point in triangle expressions
				# http://answers.yahoo.com/question/index?qid=20111103091813AA1jksL
				pi = copyBoundary[i]
				ep = (bi.x-bp.x)*(pi.y-bp.y)-(bi.y-bp.y)*(pi.x-bp.x)
				ei = (bn.x-bi.x)*(pi.y-bi.y)-(bn.y-bi.y)*(pi.x-bi.x)
				en = (bp.x-bn.x)*(pi.y-bn.y)-(bp.y-bn.y)*(pi.x-bn.x)
				
				# This only tests if the point is inside the triangle (no edge / vertex test)
				if (ep < 0 and ei < 0 and en < 0) or (ep > 0 and ei > 0 and en > 0):
					inside = True;
					break
			
			# No vertices in the "ear", add a triangle and remove bi
			if not inside:
				outSurface.append(Triangle(bp, bi, bn))
				copyBoundary.pop(index)
		index = (index+1)%len(copyBoundary)
	return outSurface

ticks = time.time()
for i in range(Community.num):
	aci = allCommunities[i]
	
	# Move the boundary to the origin before creating models (floating point accuracy)
	bLen = len(aci.boundary)
	for j in range(bLen): aci.boundary[j] -= aci.polyCenter
	
	# Given clockwise vertices defining a boundary, create a surface
	triangleSurface = triangularize(aci.boundary)
	
	# Use the triangle surface to create a model
	mgSt = ModelGeometry.create(aci.name+" Surface Top")
	tsLen = len(triangleSurface)
	for j in range(tsLen):
		jTri = triangleSurface[j]
		mgSt.addVertex(Vector3(jTri.a.x, jTri.a.y, 0.0))
		mgSt.addVertex(Vector3(jTri.c.x, jTri.c.y, 0.0))
		mgSt.addVertex(Vector3(jTri.b.x, jTri.b.y, 0.0))
	mgSt.addPrimitive(PrimitiveType.Triangles, 0, (tsLen<<1)+tsLen)
	getSceneManager().addModel(mgSt)
	
	# Give the top surface model to allCommunities[i]
	allCommunities[i].surfaceTop = StaticObject.create(aci.name+" Surface Top")
	
	# Use allCommunities[i].boundary to create a model
	mgSs = ModelGeometry.create(aci.name+" Surface Side")
	bLenMinusOne = bLen-1
	for j in range(bLenMinusOne):
		mgSs.addVertex(Vector3(aci.boundary[j].x, aci.boundary[j].y, 0.0))
		mgSs.addVertex(Vector3(aci.boundary[j].x, aci.boundary[j].y, 10.0))
	mgSs.addVertex(Vector3(aci.boundary[0].x, aci.boundary[0].y, 0.0))
	mgSs.addVertex(Vector3(aci.boundary[0].x, aci.boundary[0].y, 10.0))
	mgSs.addPrimitive(PrimitiveType.TriangleStrip, 0, bLen<<1)
	
	# Give the side surface model to allCommunities[i]
	getSceneManager().addModel(mgSs)
	allCommunities[i].surfaceSide = StaticObject.create(aci.name+" Surface Side")
	allCommunities[i].surfaceSide.setPosition(Vector3(aci.polyCenter.x, aci.polyCenter.y, 0.0))
	
	# Move each allCommunities[i].boundary[j] back to where UTM placed it
	for j in range(bLen): aci.boundary[j] += aci.polyCenter
	print "Added "+aci.name+"'s, top surface vertices = "+str(tsLen)+", side surface vertices = "+str(bLen<<1)
print "Time to add all community geometry = "+str(time.time()-ticks)

# --------------------------------------------------------------------------
# Create geometry for each element of allCTA[]'s line segments and stations
# --------------------------------------------------------------------------
ticks = time.time()
for i in range(CTA.num):
	# Height is based on index to avoid overlap of CTA lines and stations
	iHeight = 10.0+i*20.0
	
	lsLen = len(allCTA[i].lineSegments)
	for j in range(lsLen):
		ls = LineSet.create()
		vLen = len(allCTA[i].lineSegments[j])-1
		for k in range(vLen):
			start = allCTA[i].lineSegments[j][k]
			end = allCTA[i].lineSegments[j][k+1]
			
			# Add geometry for a line segment k
			l = ls.addLine()
			l.setStart(Vector3(start.x, start.y, iHeight))
			l.setEnd(Vector3(end.x, end.y, iHeight))
			l.setThickness(20.0)
			
			# Fill in space between line segments k and k+1
			s = SphereShape.create(10.0, 1)
			s.setPosition(Vector3(end.x, end.y, iHeight))
			s.setEffect(allCTA[i].color)
		ls.setEffect(allCTA[i].color)
	
	sLen = len(allCTA[i].stations)
	for j in range(sLen):
		b = BoxShape.create(60.0, 60.0, 20.0)
		b.setPosition(Vector3(allCTA[i].stations[j].loc.x, allCTA[i].stations[j].loc.y, iHeight))
		b.setEffect(allCTA[i].color)
print "Time to add all CTA geometry = "+str(time.time()-ticks)

# -------------------------------
# Update community area geometry
# -------------------------------
def updateCAHeight():
	# Count the number of crimes within the time index range
	crimeCounts = [0]*Community.num
	for i in range(Community.num):
		aci = allCommunities[i]
		cLen = len(aci.crimes)
		for j in range(cLen):
			ccj = aci.crimes[j]
			if ccj.year >= crimeStartYear and ccj.month >= crimeStartMonth and ccj.year <= crimeEndYear and ccj.month <= crimeEndYear:
				if crimeTypeSelection == "All" or ccj.kind == crimeTypeSelection: crimeCounts[i] += 1
	
	# Update the community area geometry
	maxCrimes = max(crimeCounts)
	if maxCrimes == 0:
		for i in range(Community.num):
			aci = allCommunities[i]
			allCommunities[i].surfaceTop.setPosition(Vector3(aci.polyCenter.x, aci.polyCenter.y, 10.0))
			allCommunities[i].surfaceTop.setEffect("colored -d #ffffffc0 -t")
			allCommunities[i].surfaceSide.setScale(1.0, 1.0, 1.0)
			allCommunities[i].surfaceSide.setEffect("colored -d #ffffffc0 -t")
	else:
		sqrtMC = math.sqrt(maxCrimes)
		for i in range(Community.num):
			sqrtCC = crimeCounts[i] = 0.0 if crimeCounts[i] == 0 else math.sqrt(crimeCounts[i])
			percent = sqrtCC/sqrtMC
			polyHeight = percent*990.0+10.0
			
			hexVal = hex(255-int(percent*255))[2:4]
			if len(hexVal) == 1: hexVal = "0"+hexVal
			hexColor = "colored -d #ff"+hexVal+hexVal+"c0 -t"
			
			aci = allCommunities[i]
			allCommunities[i].surfaceTop.setPosition(Vector3(aci.polyCenter.x, aci.polyCenter.y, polyHeight))
			allCommunities[i].surfaceTop.setEffect(hexColor)
			allCommunities[i].surfaceSide.setScale(1.0, 1.0, polyHeight/10.0)
			allCommunities[i].surfaceSide.setEffect(hexColor)
updateCAHeight()

# -----------------------------
# Add "Select Map" to the menu
# -----------------------------
menuSelectMap = mainMenu.getMainMenu().addSubMenu("Select Map")
osgButton = menuSelectMap.addButton("OSG City", "osgCity.setVisible(True); googleCity.setVisible(False); yahooCity.setVisible(False)")
osgButton.getButton().setCheckable(True)
osgButton.getButton().setRadio(True)
googleButton = menuSelectMap.addButton("Google City", "osgCity.setVisible(False); googleCity.setVisible(True); yahooCity.setVisible(False)")
googleButton.getButton().setCheckable(True)
googleButton.getButton().setRadio(True)
yahooButton = menuSelectMap.addButton("Yahoo City", "osgCity.setVisible(False); googleCity.setVisible(False); yahooCity.setVisible(True)")
yahooButton.getButton().setCheckable(True)
yahooButton.getButton().setRadio(True)

# Initialize OSG as selected
osgButton.getButton().setChecked(True)
googleCity.setVisible(False)
yahooCity.setVisible(False)

# ---------------------------------------
# Add "Visit Community Area" to the menu
# ---------------------------------------
visitCAPage = 0
visitCARange = 12
visitCASlots = []

def updateVisitCAMenu(inAlter):
	global visitCAPage # Python is stupid
	
	newCAPage = visitCAPage+inAlter
	if newCAPage >= 0 and newCAPage < Community.num:
		visitCAPage = newCAPage
		
		# Update "menuVistCA" content
		for i in range(visitCARange):
			index = visitCAPage+i
			if index < Community.num:
				visitCASlots[i].setText(allCommunities[index].name)
				visitCASlots[i].setCommand("visitCA("+str(index)+")")
				visitCASlots[i].getButton().setCheckable(True)
				visitCASlots[i].getButton().setRadio(True)
				visitCASlots[i].getButton().setChecked(False)
			else:
				visitCASlots[i].setText("-")
				visitCASlots[i].setCommand("print 'Nothing to visit'")
				visitCASlots[i].getButton().setCheckable(False)
				visitCASlots[i].getButton().setRadio(False)
	print "Viewing page "+str(visitCAPage/visitCARange)

def visitCA(inIndex):
	camOffset = allCommunities[inIndex].surfaceTop.getBoundRadius()*2.0
	if (camOffset < 1500.0): camOffset = 1500.0
	cam.setPosition(allCommunities[inIndex].polyCenter.x, allCommunities[inIndex].polyCenter.y-camOffset, camOffset)
	cam.setPitchYawRoll(Vector3(math.pi*0.25, 0.0, 0.0))

# "Overview" is always an option
menuVisitCA = mainMenu.getMainMenu().addSubMenu("Visit Community Area")
overviewButton = menuVisitCA.addButton("Overview", "seeCity()")
overviewButton.getButton().setCheckable(True)
overviewButton.getButton().setRadio(True)
overviewButton.getButton().setChecked(True)

# "Previous" is always an option
menuVisitCA.addButton("Previous", "updateVisitCAMenu(-visitCARange)")

# Initialize visitCASlots[i] with the first visitCARange of allCommunities[i].name
for i in range(visitCARange):
	visitCASlots.append(menuVisitCA.addButton(allCommunities[i].name, "visitCA("+str(i)+")"))
	visitCASlots[i].getButton().setCheckable(True)
	visitCASlots[i].getButton().setRadio(True)

# "Next" is always an option
menuVisitCA.addButton("Next", "updateVisitCAMenu(visitCARange)")

# -----------------------------
# Add "Crime Type" to the menu
# -----------------------------
# Make a crime type array quickly by using a dictionary
ticks = time.time()
crimeTypeArr = []
crimeTypeDict = {}
for i in range(Community.num):
	aci = allCommunities[i]
	cLen = len(aci.crimes)
	for j in range(cLen):
		ccjk = aci.crimes[j].kind
		if ccjk not in crimeTypeDict:
			crimeTypeDict[ccjk] = None
			crimeTypeArr.append(ccjk)
del crimeTypeDict
print "Time to make the crime type array = "+str(time.time()-ticks)

# Sort the crime type array lexicographically
crimeTypeArr.sort()
print crimeTypeArr

crimeTypePage = 0
crimeTypeRange = 12
crimeTypeSlots = []

def setCrimeType(inIndex):
	global crimeTypeSelection # Python is stupid
	
	if inIndex == -1: crimeTypeSelection = "All"
	else: crimeTypeSelection = crimeTypeArr[inIndex]
	print crimeTypeSelection
	updateCAHeight()

def updateCrimeTypeMenu(inAlter):
	global crimeTypePage # Python is stupid
	
	newCTPage = crimeTypePage+inAlter
	ctLen = len(crimeTypeArr)
	if newCTPage >= 0 and newCTPage < ctLen:
		crimeTypePage = newCTPage
		
		# Update "menuCrimeType" content
		for i in range(crimeTypeRange):
			index = crimeTypePage+i
			if index < ctLen:
				crimeTypeSlots[i].setText(crimeTypeArr[index])
				crimeTypeSlots[i].setCommand("setCrimeType("+str(index)+")")
				crimeTypeSlots[i].getButton().setCheckable(True)
				crimeTypeSlots[i].getButton().setRadio(True)
				crimeTypeSlots[i].getButton().setChecked(False)
			else:
				crimeTypeSlots[i].setText("-")
				crimeTypeSlots[i].setCommand("print 'Not a crime type'")
				crimeTypeSlots[i].getButton().setCheckable(False)
				crimeTypeSlots[i].getButton().setRadio(False)
	print "Viewing page "+str(crimeTypePage/crimeTypeRange)

# "All Crimes" is always an option
menuCrimeType = mainMenu.getMainMenu().addSubMenu("Crime Type")
allCrimeButton = menuCrimeType.addButton("All Crimes", "setCrimeType("+str(-1)+")")
allCrimeButton.getButton().setCheckable(True)
allCrimeButton.getButton().setRadio(True)
allCrimeButton.getButton().setChecked(True)

# "Previous" is always an option
menuCrimeType.addButton("Previous", "updateCrimeTypeMenu(-crimeTypeRange)")

# Initialize crimeTypeSlots[i] with the first crimeTypeRange of crimeTypeArr[i]
for i in range(crimeTypeRange):
	crimeTypeSlots.append(menuCrimeType.addButton(crimeTypeArr[i], "setCrimeType("+str(i)+")"))
	crimeTypeSlots[i].getButton().setCheckable(True)
	crimeTypeSlots[i].getButton().setRadio(True)

# "Next" is always an option
menuCrimeType.addButton("Next", "updateCrimeTypeMenu(crimeTypeRange)")

# --------------------------------------------------------
# Add "Time Start Index" and "Time End Index" to the menu
# --------------------------------------------------------
monthList = ["January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"]

def setCrimeStart(inYear, inMonth):
	global crimeStartYear, crimeStartMonth, crimeEndYear, crimeEndMonth # Python is stupid
	
	crimeStartYear = inYear
	crimeStartMonth = inMonth
	if crimeStartYear > crimeEndYear:
		crimeEndYear = crimeStartYear
		crimeEndMonth = crimeStartMonth
	elif crimeStartYear == crimeEndYear and crimeStartMonth > crimeEndMonth: crimeEndMonth = crimeStartMonth
	
	print "Crime start = "+str(crimeStartYear)+", "+str(crimeStartMonth)+"; Crime end = "+str(crimeEndYear)+", "+str(crimeEndMonth)
	updateCAHeight()

def setCrimeEnd(inYear, inMonth):
	global crimeStartYear, crimeStartMonth, crimeEndYear, crimeEndMonth # Python is stupid
	
	crimeEndYear = inYear
	crimeEndMonth = inMonth
	if crimeEndYear < crimeStartYear:
		crimeStartYear = crimeEndYear
		crimeStartMonth = crimeEndMonth
	elif crimeEndYear == crimeStartYear and crimeEndMonth < crimeStartMonth: crimeStartMonth = crimeEndMonth
	
	print "Crime start = "+str(crimeStartYear)+", "+str(crimeStartMonth)+"; Crime end = "+str(crimeEndYear)+", "+str(crimeEndMonth)
	updateCAHeight()

# Time index menus are trees where leaves are buttons
menuTimeStart = mainMenu.getMainMenu().addSubMenu("Time Start Index")
menuTimeEnd = mainMenu.getMainMenu().addSubMenu("Time End Index")

for i in range(2001, 2014):
	currYear = menuTimeStart.addSubMenu(str(i))
	for j in range(1, 13):
		currYear.addButton(monthList[j-1], "setCrimeStart("+str(i)+","+str(j)+")")

	currYear = menuTimeEnd.addSubMenu(str(i))
	for j in range(1, 13):
		currYear.addButton(monthList[j-1], "setCrimeEnd("+str(i)+","+str(j)+")")
